import { Construct } from "constructs";
/**
 * Policy substitutions provided as key-value pairs.
 */
export interface PolicyMapping {
    /**
     * Name of substitution variable, e.g., `region` or `account`.
     */
    readonly name: string;
    /**
     * Value of substitution variable, e.g., `us-east-1` or `12345689012`
     */
    readonly value: string;
}
/**
 * Properties for defining an AWS IoT thing, AWS IoT certificate, and AWS IoT
 * policy.
 * @summary The properties for the IotThingCertificatePolicy class.
 */
export interface IotThingCertificatePolicyProps {
    /**
     * Name of AWS IoT thing to create.
     *
     * @default - None
     */
    readonly thingName: string;
    /**
     * Name of the AWS IoT Core policy to create.
     *
     * @default - None
     */
    readonly iotPolicyName: string;
    /**
     * The AWS IoT policy in JSON format to be created and attached to the certificate.
     * This is a JSON string that uses [mustache-compatible](https://handlebarsjs.com/guide/)
     * template substitution to create the AWS IoT policy.
     *
     * @default - None
     */
    readonly iotPolicy: string;
    /**
     * Optional: A `PolicyMapping` object of parameters and values to be replaced if a
     * [mustache-compatible](https://handlebarsjs.com/guide/) template is
     * provided as the `iotPolicy` (see example).
     * For each matching parameter in the policy template, the value
     * will be used. If not provided, only the `{{thingname}}` mapping will be available for the
     * `iotPolicy` template.
     *
     * @default - None
     */
    readonly policyParameterMapping?: PolicyMapping[];
    /**
     * Selects RSA or ECC private key and certificate generation. If not provided, `RSA` will be used.
     *
     * @default - RSA
     */
    readonly encryptionAlgorithm?: string;
}
/**
 * Creates and associates an AWS IoT thing, AWS IoT certificate, and AWS IoT policy.
 * It attaches the certificate to the thing and policy, and then stores the certificate
 * and private key in AWS Systems Manager Parameter Store parameters for reference
 * outside of the CloudFormation stack or by other constructs.
 *
 * Use this construct to create and delete a thing, certificate (principal), and IoT policy for
 * testing or other singular uses. **Note:** Destroying this stack will fully detach and delete
 * all created IoT resources.
 *
 * @summary Creates and associates an AWS IoT thing, certificate and policy.
 */
export declare class IotThingCertificatePolicy extends Construct {
    /**
     * Arn of created AWS IoT Thing.
     */
    readonly thingArn: string;
    /**
     * Arn of created AWS IoT Policy.
     */
    readonly iotPolicyArn: string;
    /**
     * Arn of created AWS IoT Certificate.
     */
    readonly certificateArn: string;
    /**
     * Fully qualified name in AWS Systems Manager Parameter Store of the certificate in `PEM` format.
     */
    readonly certificatePemParameter: string;
    /**
     * Fully qualified name in AWS Systems Manager Parameter Store of the certificate's private key in `PEM` format.
     */
    readonly privateKeySecretParameter: string;
    /**
     * Fully qualified domain name of the AWS IoT Core data plane endpoint specific to this AWS account and AWS region.
     */
    readonly dataAtsEndpointAddress: string;
    /**
     * Fully qualified domain name of the AWS IoT Credential provider endpoint specific to this AWS account and AWS region.
     */
    readonly credentialProviderEndpointAddress: string;
    private customResourceName;
    /**
     * @summary Constructs a new instance of the `IotThingCertificatePolicy` class.
     * @param {cdp.App} scope Represents the scope for all the resources.
     * @param {string} id This is a scope-unique id.
     * @param {IotThingCertificatePolicyProps} props User provided properties for the construct.
     * @since 2.138.0
     */
    constructor(scope: Construct, id: string, props: IotThingCertificatePolicyProps);
}
